@@ -27,7 +27,22 @@ envelopes.envelope
 This module contains the Envelope class.
 """
 
-from email import Encoders
+import sys
+
+if sys.version_info[0] == 2:
+    from email import Encoders as email_encoders
+elif sys.version_info[0] == 3:
+    from email import encoders as email_encoders
+    basestring = str
+
+    def unicode(_str, _charset):
+        return str(_str.encode(_charset), _charset)
+else:
+    raise RuntimeError('Unsupported Python version: %d.%d.%d' % (
+        sys.version_info[0], sys.version_info[1], sys.version_info[2]
+    ))
+
+from email.header import Header
 from email.mime.base import MIMEBase
 from email.mime.multipart import MIMEMultipart
 from email.mime.application import MIMEApplication
@@ -36,14 +51,15 @@ from email.mime.image import MIMEImage
 from email.mime.text import MIMEText
 import mimetypes
 import os
+import re
 
 from .conn import SMTP
+from .compat import encoded
 
 
 class MessageEncodeError(Exception):
     pass
 
-
 class Envelope(object):
     """
     The Envelope class.
@@ -64,13 +80,14 @@ class Envelope(object):
     :param subject: message subject
     :param html_body: optional HTML part of the message
     :param text_body: optional plain text part of the message
-    :param cc_addrs: optional list of CC address
-    :param bcc_addrs: optional list of BCC address
+    :param cc_addr: optional single CC address or list of CC addresses
+    :param bcc_addr: optional single BCC address or list of BCC addresses
     :param headers: optional dictionary of headers
     :param charset: message charset
     """
 
     ADDR_FORMAT = '%s <%s>'
+    ADDR_REGEXP = re.compile(r'^(.*) <([^@]+@[^@]+)>$')
 
     def __init__(self, to_addr=None, from_addr=None, subject=None,
                  html_body=None, text_body=None, cc_addr=None, bcc_addr=None,
@@ -94,12 +111,18 @@ class Envelope(object):
             self._parts.append(('text/html', html_body, charset))
 
         if cc_addr:
-            self._cc = cc_addr
+            if isinstance(cc_addr, list):
+                self._cc = cc_addr
+            else:
+                self._cc = [cc_addr]
         else:
             self._cc = []
 
         if bcc_addr:
-            self._bcc = bcc_addr
+            if isinstance(bcc_addr, list):
+                self._bcc = bcc_addr
+            else:
+                self._bcc = [bcc_addr]
         else:
             self._bcc = []
 
@@ -112,6 +135,13 @@ class Envelope(object):
 
         self._addr_format = unicode(self.ADDR_FORMAT, charset)
 
+    def __repr__(self):
+        return u'<Envelope from="%s" to="%s" subject="%s">' % (
+            self._addrs_to_header([self._from]),
+            self._addrs_to_header(self._to),
+            self._subject
+        )
+
     @property
     def to_addr(self):
         """List of ``To`` addresses."""
@@ -175,7 +205,7 @@ class Envelope(object):
 
         if len(addr_tuple) == 2 and addr_tuple[1]:
             addr = self._addr_format % (
-                addr_tuple[1] or '',
+                self._header(addr_tuple[1] or ''),
                 addr_tuple[0] or ''
             )
         elif addr_tuple[0]:
@@ -203,47 +233,63 @@ class Envelope(object):
                 continue
 
             if isinstance(addr, basestring):
-                _addrs.append(addr)
+                if self._is_ascii(addr):
+                    _addrs.append(self._encoded(addr))
+                else:
+                    # these headers need special care when encoding, see:
+                    #   http://tools.ietf.org/html/rfc2047#section-8
+                    # Need to break apart the name from the address if there are
+                    # non-ascii chars
+                    m = self.ADDR_REGEXP.match(addr)
+                    if m:
+                        t = (m.group(2), m.group(1))
+                        _addrs.append(self._addr_tuple_to_addr(t))
+                    else:
+                        # What can we do? Just pass along what the user gave us and hope they did it right
+                        _addrs.append(self._encoded(addr))
             elif isinstance(addr, tuple):
                 _addrs.append(self._addr_tuple_to_addr(addr))
             else:
                 self._raise(MessageEncodeError,
                             '%s is not a valid address' % str(addr))
 
-        _header = unicode(',', self._charset).join(_addrs)
+        _header = ','.join(_addrs)
         return _header
 
     def _raise(self, exc_class, message):
-        if isinstance(message, unicode):
-            msg = message
-        else:
-            msg = unicode(message, self._charset)
+        raise exc_class(self._encoded(message))
 
-        raise exc_class(msg.encode(self._charset))
+    def _header(self, _str):
+        if self._is_ascii(_str):
+            return _str
+        return Header(_str, self._charset).encode()
+
+    def _is_ascii(self, _str):
+        return all(ord(c) < 128 for c in _str)
+
+    def _encoded(self, _str):
+        return encoded(_str, self._charset)
 
     def to_mime_message(self):
         """Returns the envelope as
         :py:class:`email.mime.multipart.MIMEMultipart`."""
         msg = MIMEMultipart('alternative')
-        msg['Subject'] = (self._subject or '').encode(self._charset)
+        msg['Subject'] = self._header(self._subject or '')
 
-        msg['From'] = self._addrs_to_header([self._from]).encode(self._charset)
-        msg['To'] = self._addrs_to_header(self._to).encode(self._charset)
+        msg['From'] = self._encoded(self._addrs_to_header([self._from]))
+        msg['To'] = self._encoded(self._addrs_to_header(self._to))
 
         if self._cc:
-            msg['CC'] = self._addrs_to_header(self._cc).encode(self._charset)
-
-        if self._bcc:
-            msg['BCC'] = self._addrs_to_header(self._bcc).encode(self._charset)
+            msg['CC'] = self._addrs_to_header(self._cc)
 
         if self._headers:
-            for key, value in self._headers.iteritems():
-                msg[key] = value.encode(self._charset)
+            for key, value in self._headers.items():
+                msg[key] = self._header(value)
 
         for part in self._parts:
             type_maj, type_min = part[0].split('/')
             if type_maj == 'text' and type_min in ('html', 'plain'):
-                msg.attach(MIMEText(part[1].encode(self._charset), type_min))
+                msg.attach(MIMEText(part[1], type_min, self._charset))
             else:
                 msg.attach(part[1])
 
@@ -265,9 +311,9 @@ class Envelope(object):
 
             part = MIMEBase(type_maj, type_min)
             part.set_payload(part_data)
-            Encoders.encode_base64(part)
+            email_encoders.encode_base64(part)
 
-            part_filename = os.path.basename(file_path.encode(self._charset))
+            part_filename = os.path.basename(self._encoded(file_path))
             part.add_header('Content-Disposition', 'attachment; filename="%s"'
                             % part_filename)
 
